/**
 * Copyright (c) 2011-2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.logger;

import java.util.logging.Logger;

import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.MessageExchange;

import org.ow2.petals.commons.log.FlowAttributes;
import org.ow2.petals.commons.log.Level;

import com.ebmwebsourcing.easycommons.lang.StringHelper;

/**
 * Helpers for logging MONIT traces.
 * 
 * Normally the {@link FlowAttributes} parameters should never be <code>null</code>, but to avoid problems, it will
 * still work if it is the case.
 * 
 * @author vnoel
 *
 */
public class StepLogHelper {

    private static final String ERROR_MESSAGE_LINE_SEPARATOR = " - ";

    public static final String BUSINESS_ERROR_MESSAGE = "A business error occurs";

    public static final String TECHNICAL_ERROR_MESSAGE_PATTERN = "A technical error occurs: %s";

    public static final String UNKNOWN_ERROR_MESSAGE = "A unknown error occurs (%s)";

    private StepLogHelper() {
        // NOP
    }

    /**
     * Logs the start of a ProvideStep, based on the content of the exchange.
     * 
     * @param logger
     *            The logger used to log the expected MONIT trace
     * @param exchange
     *            The exchange from which we extract the end type: normal end or failure
     * @param flowAttributes
     *            the current flow attributes (and not those potentially in the exchange). If {@code null}, flow
     *            attributes will be extracted from the exchange.
     * @param previous
     *            The flow attributes of the previous step if exists. Can be {@code null}.
     */
    public static void addMonitStartTrace(final Logger logger, final MessageExchange exchange,
            final FlowAttributes current, final FlowAttributes previous) {
        final String flowInstanceId;
        final String flowStepId;
        if (current != null) {
            flowInstanceId = current.getFlowInstanceId();
            flowStepId = current.getFlowStepId();
        } else {
            flowInstanceId = null;
            flowStepId = null;
        }

        final String prevFlowStepId;
        if (previous != null) {
            assert flowInstanceId != null && flowInstanceId.equals(previous.getFlowInstanceId());
            prevFlowStepId = previous.getFlowStepId();
        } else {
            prevFlowStepId = null;
        }

        final String interfaceName = StringHelper.nonNullValue(exchange.getInterfaceName());
        final String serviceName = StringHelper.nonNullValue(exchange.getService());
        final String operationName = StringHelper.nonNullValue(exchange.getOperation());
        final String endpointName = StringHelper
                .nonNullValue(exchange.getEndpoint() != null ? exchange.getEndpoint().getEndpointName() : null);

        final AbstractFlowLogData flowLogData = new ProvideFlowStepBeginLogData(flowInstanceId, flowStepId,
                interfaceName, serviceName, operationName, endpointName, prevFlowStepId, exchange);

        logger.log(Level.MONIT, "", flowLogData);
    }

    /**
     * Logs the end of a ProvideStep, based on the content of the exchange.
     * 
     * @param logger
     *            The logger used to log the expected MONIT trace
     * @param exchange
     *            The exchange from which we extract the end type: normal end or failure
     * @param flowAttributes
     *            the current flow attributes (and not those potentially in the exchange)
     */
    public static void addMonitEndOrFailureTrace(final Logger logger, final MessageExchange exchange,
            final FlowAttributes flowAttributes) {
        final String flowInstanceId;
        final String flowStepId;
        if (flowAttributes != null) {
            flowInstanceId = flowAttributes.getFlowInstanceId();
            flowStepId = flowAttributes.getFlowStepId();
        } else {
            flowInstanceId = null;
            flowStepId = null;
        }

        final AbstractFlowLogData flowLogData;
        final boolean isError = ExchangeStatus.ERROR.equals(exchange.getStatus());
        final boolean isFault = exchange.getFault() != null;

        if (!isError && !isFault) {
            flowLogData = new ProvideFlowStepEndLogData(flowInstanceId, flowStepId, exchange);
        } else {
            final String failureMessage = extractFailureMessage(exchange);
            flowLogData = new ProvideFlowStepFailureLogData(flowInstanceId, flowStepId, failureMessage, exchange);
        }

        logger.log(Level.MONIT, "", flowLogData);
    }
    
    public static void addMonitExtEndOrFailureTrace(final Logger logger, final MessageExchange exchange,
            final FlowAttributes flowAttributes, final boolean incoming) {

        final boolean isError = ExchangeStatus.ERROR.equals(exchange.getStatus());
        final boolean isFault = exchange.getFault() != null;

        if (!isError && !isFault) {
            addMonitExtEndTrace(logger, flowAttributes, incoming);
        } else {
            final String failureMessage = extractFailureMessage(exchange);
            addMonitExtFailureTrace(logger, flowAttributes, failureMessage, incoming);
        }
    }

    /**
     * Logs an EndStep (typically used by BCs)
     * 
     * @param logger
     *            The logger used to log the expected MONIT trace
     * @param flowAttributes
     *            the current flow attributes (and not those potentially in the exchange)
     * @param incoming
     *            if {@code true} a ConsumeExtFlowStepEnd will be logged (it's about an incoming {@link Exchange}), else
     *            a ProvideExtFlowStepEnd will be logged (it's about an outgoing {@link Exchange}).
     */
    public static void addMonitExtEndTrace(final Logger logger, final FlowAttributes flowAttributes,
            final boolean incoming) {
        final String flowInstanceId;
        final String flowStepId;
        if (flowAttributes != null) {
            flowInstanceId = flowAttributes.getFlowInstanceId();
            flowStepId = flowAttributes.getFlowStepId();
        } else {
            flowInstanceId = null;
            flowStepId = null;
        }

        final AbstractFlowLogData flowLogData;
        if (incoming) {
            flowLogData = new ConsumeExtFlowStepEndLogData(flowInstanceId, flowStepId);
        } else {
            flowLogData = new ProvideExtFlowStepEndLogData(flowInstanceId, flowStepId);
        }

        logger.log(Level.MONIT, "", flowLogData);
    }

    /**
     * Logs a FailureStep (typically used by BCs)
     * 
     * @param logger
     *            The logger used to log the expected MONIT trace
     * @param flowAttributes
     *            the current flow attributes (and not those potentially in the exchange)
     * @param error
     *            The cause of failure
     * @param incoming
     *            if {@code true} a ConsumeExtFlowStepFailure will be logged (it's about an incoming {@link Exchange}),
     *            else a ProvideExtFlowStepFailure will be logged (it's about an outgoing {@link Exchange}).
     */
    public static void addMonitExtFailureTrace(final Logger logger, final FlowAttributes flowAttributes,
            final Exception error, final boolean incoming) {
        addMonitExtFailureTrace(logger, flowAttributes, extractFailureMessage(error), incoming);
    }

    /**
     * Logs a FailureStep (typically used by BCs)
     * 
     * @param logger
     *            The logger used to log the expected MONIT trace
     * @param flowAttributes
     *            the current flow attributes (and not those potentially in the exchange)
     * @param failureMessage
     *            The failure message
     * @param incoming
     *            if {@code true} a ConsumeExtFlowStepFailure will be logged (it's about an incoming {@link Exchange}),
     *            else a ProvideExtFlowStepFailure will be logged (it's about an outgoing {@link Exchange}).
     */
    public static void addMonitExtFailureTrace(final Logger logger, final FlowAttributes flowAttributes,
            final String failureMessage, final boolean incoming) {

        final String msg = failureMessage.replace("\n", ERROR_MESSAGE_LINE_SEPARATOR);

        final String flowInstanceId;
        final String flowStepId;
        if (flowAttributes != null) {
            flowInstanceId = flowAttributes.getFlowInstanceId();
            flowStepId = flowAttributes.getFlowStepId();
        } else {
            flowInstanceId = null;
            flowStepId = null;
        }

        final AbstractFlowLogData flowLogData;
        if (incoming) {
            flowLogData = new ConsumeExtFlowStepFailureLogData(flowInstanceId, flowStepId, msg);
        } else {
            flowLogData = new ProvideExtFlowStepFailureLogData(flowInstanceId, flowStepId, msg);
        }

        logger.log(Level.MONIT, "", flowLogData);
    }

    private static String extractFailureMessage(final MessageExchange exchange) {

        final boolean isError = ExchangeStatus.ERROR.equals(exchange.getStatus());

        final String failureMessage;
        if (isError) {
            final Exception error = exchange.getError();
            if (error != null) {
                failureMessage = extractFailureMessage(error);
            } else {
                failureMessage = String.format(UNKNOWN_ERROR_MESSAGE, "no error set");
            }
        } else {
            failureMessage = BUSINESS_ERROR_MESSAGE;
        }

        return failureMessage;
    }


    private static String extractFailureMessage(final Exception error) {
        final String failureMessage;
        final String message = error.getMessage();
        if (message != null) {
            failureMessage = message.replace("\n", ERROR_MESSAGE_LINE_SEPARATOR);
        } else if (error.getStackTrace().length > 0) {
            failureMessage = String.format(UNKNOWN_ERROR_MESSAGE, error.getStackTrace()[0].toString());
        } else {
            failureMessage = String.format(UNKNOWN_ERROR_MESSAGE, error.getClass().getName());
        }
        return failureMessage;
    }
    
}
